@@ -120,7 +120,7 @@ if (MIME::Charset::USE_ENCODE) {
 #------------------------------
 
 ### The package version, both in 1.23 style *and* usable by MakeMaker:
-$VERSION = '1.009';
+$VERSION = '1.010';
 
 ### Public Configuration Attributes
 $Config = {
@@ -142,6 +142,7 @@ my $PRINTABLE = "\\x21-\\x7E";
 my $NONPRINT = qr{[^$PRINTABLE]}; # Improvement: Unicode support.
 my $UNSAFE = qr{[^\x01-\x20$PRINTABLE]};
 my $WIDECHAR = qr{[^\x00-\xFF]};
+my $ASCIITRANS = qr{^(?:HZ-GB-2312|UTF-7)$}i;
 
 #------------------------------
 
@@ -180,9 +181,9 @@ sub _encode_B {
 #     Improvement by this module: Spaces are escaped by ``_''.
 sub _encode_Q {
     my $str = shift;
-    # $str =~ s{([_\?\=$NONPRINT])}{sprintf("=%02X", ord($1))}eog;
-    $str =~ s{(\x20)|([_?=]|$NONPRINT)}{
-	defined $1? "_": sprintf("=%02X", ord($2))
+    # Restrict characters to those listed in RFC 2047 section 5 (3)
+    $str =~ s{[^-!*+/0-9A-Za-z]}{
+	$& eq "\x20"? "_": sprintf("=%02X", ord($&))
 	}eog;
     $str;
 }
@@ -483,9 +484,17 @@ Spaces will be escaped with ``_'' by C<"Q"> encoding.
 sub encode_mimeword {
     my $word = shift;
     my $encoding = uc(shift || 'Q');          # not overridden.
-    my $charset  = uc(shift || 'ISO-8859-1'); # ditto.
+    my $charset  = shift || 'ISO-8859-1';     # ditto.
     my $language = uc(shift || "");	      # ditto.
 
+    if (ref $charset) {
+	if (is_utf8($word) or $word =~ /$WIDECHAR/) {
+	    $word = $charset->undecode($word, 0);
+	}
+	$charset = $charset->as_string;
+    } else {
+	$charset = uc($charset);
+    }
     my $encstr;
     if ($encoding eq 'Q') {
 	$encstr = &_encode_Q($word);
@@ -639,17 +648,16 @@ sub encode_mimewords  {
     my $words = shift;
     my %params = @_;
     my %Params = &_getparams(\%params,
-			     NoDefault => [qw(Charset)], # apply default later
 			     YesNo => [qw(Detect7bit Minimal)],
-			     Others => [qw(Encoding Field Folding Mapping
-					   MaxLineLen Replacement)],
+			     Others => [qw(Charset Encoding Field Folding
+					   Mapping MaxLineLen Replacement)],
 			     ToUpper => [qw(Charset Encoding Mapping
 					    Replacement)],
 			    );
-    # checks
     croak "unsupported encoding ``$Params{Encoding}''"
 	unless $Params{Encoding} =~ /^[ABQS]$/;
-    my ($fwsbrk, $fwsspc);		# Newline and following WSP
+    # newline and following WSP
+    my ($fwsbrk, $fwsspc);
     if ($Params{Folding} =~ m/^([\r\n]*)([\t ]?)$/) {
 	$fwsbrk = $1;
 	$fwsspc = $2 || " ";
@@ -657,90 +665,115 @@ sub encode_mimewords  {
 	croak sprintf "illegal folding sequence ``\\x%*v02X''", '\\x',
 		      $Params{Folding};
     }
-    # calculate some variables
+    # charset objects
     my $charsetobj = MIME::Charset->new($Params{Charset},
 					Mapping => $Params{Mapping});
+    my $ascii = MIME::Charset->new("US-ASCII", Mapping => 'STANDARD');
+    $ascii->encoder($ascii);
+    # lengths
     my $firstlinelen = $Params{MaxLineLen} -
 	($Params{Field}? length("$Params{Field}: "): 0);
     my $maxrestlen = $Params{MaxLineLen} - length($fwsspc);
+    my $UNSAFEASCII = ($maxrestlen <= 1)?
+	qr{(?: =\? )}ox:
+	qr{(?: =\? | [$PRINTABLE]{$Params{MaxLineLen}} )}ox;
 
     unless (ref($words) eq "ARRAY") {
+	my @words = ();
 	# unfolding: normalize linear-white-spaces and orphan newlines.
 	$words =~ s/(?:[\r\n]+[\t ])*[\r\n]+([\t ]|\Z)/$1? " ": ""/eg;
 	$words =~ s/[\r\n]+/ /g;
 	# split if required
 	if ($Params{Minimal} eq "YES") {
-	    my @words = map { [$_, $Params{Charset}] }
-			    split(/((?:\A|[\t ])[\t $PRINTABLE]+(?:[\t ]|\Z))/,
-				  $words);
-	    $words = \@words;
+	    my ($spc, $unsafe_last) = ('', 0);
+	    foreach my $w (split(/([\t ]+)/, $words)) {
+		next unless scalar(@words) or length($w); # skip garbage
+		if ($w =~ /[\t ]/) {
+		    $spc = $w;
+		    next;
+		}
+
+		# workaround for ``ASCII transformation'' charsets
+		my $u = $w;
+		if ($charsetobj->as_string =~ /$ASCIITRANS/ and $u =~ /[+~]/) {
+		    if ($charsetobj->decoder) {
+			$u = $charsetobj->decode($u);
+		    } elsif (!MIME::Charset::USE_ENCODE) { # for pre-Encode env.
+			$u = "X$u";
+		    } else { # NOTREACHED
+			croak __PACKAGE__.": Bug in encode_mimewords";
+		    }
+		}
+		if (scalar(@words)) {
+		    if (($w =~ /$NONPRINT|$UNSAFEASCII/ or $u ne $w) xor
+			$unsafe_last) {
+			if ($unsafe_last) {
+			    push @words, $spc.$w;
+			} else {
+			    $words[-1] .= $spc;
+			    push @words, $w;
+			}
+			$unsafe_last = not $unsafe_last;
+		    } else {
+			$words[-1] .= $spc.$w;
+		    }
+		} else {
+		    push @words, $spc.$w;
+		    $unsafe_last =
+			($w =~ /$NONPRINT|$UNSAFEASCII/ or $u ne $w);
+		}
+		$spc = '';
+	    }
+	    if ($spc) {
+		if (scalar(@words)) {
+		    $words[-1] .= $spc;
+		} else { # only WSPs
+		    push @words, $spc;
+		}
+	    }
 	} else {
-	    $words = [[$words, $Params{Charset}]];
+	    @words = ($words);
 	}
+	$words = [map { [$_, $Params{Charset}] } @words];
     }
 
     # Translate / concatenate words.
     my @triplets;
     foreach (@$words) {
 	my ($s, $cset) = @$_;
-	my $csetobj = MIME::Charset->new($cset, Mapping => $Params{Mapping});
-	my $enc;
-
 	next unless length($s);
-
-	# Unicode string should be encoded by given charset.
-	# Unsupported charset will be fallbacked to UTF-8.
-	if (is_utf8($s) or $s =~ $WIDECHAR) {
-	    unless ($csetobj->decoder) {
-		if ($s !~ $UNSAFE) {
-		    $cset = "US-ASCII";
-		} elsif ($Params{Replacement} =~ /^(CROAK|STRICT)$/) {
-		    croak "unsupported charset ``$cset''";
-		} else {
-		    $cset = "UTF-8";
-		}
-	    }
-	    $csetobj = MIME::Charset->new($cset, Mapping => $Params{Mapping});
-	    if ($Params{Replacement} =~ /^(CROAK|STRICT)$/) {
-		$s = $csetobj->undecode($s, FB_CROAK());
-	    } else {
-		$s = $csetobj->undecode($s, 0);
+	my $csetobj = MIME::Charset->new($cset || "",
+					 Mapping => $Params{Mapping});
+	# determine charset and encoding
+	# try defaults only if 7-bit charset detection is not required
+	my $enc;
+	my $obj = $csetobj;
+	unless ($obj->as_string) {
+	    if ($Params{Encoding} ne "A" or $Params{Detect7bit} eq "NO" or
+		$s =~ /$UNSAFE/) {
+		$obj = $charsetobj;
 	    }
 	}
-
-	# Determine charset and encoding.
-	if ($Params{Encoding} eq "A") {
-	    my $obj = $cset? $csetobj: $charsetobj;
-	    ($s, $cset, $enc) =
-		$obj->header_encode($s,
-				    Detect7bit => $Params{Detect7bit},
-				    Replacement => $Params{Replacement});
-	    if ($cset eq "8BIT") {
-		$cset = $Config->{Charset};
-		$csetobj = MIME::Charset->new($cset,
-					      Mapping => $Params{Mapping});
-		$enc = $csetobj->header_encoding;
-	    } else {
-		$csetobj = MIME::Charset->new($cset,
-					      Mapping => $Params{Mapping});
-	    }
-	} elsif (!$cset or $cset ne "US-ASCII") {
-	    $cset ||= $Params{Charset} || $Config->{Charset};
-	    $csetobj = MIME::Charset->new($cset, Mapping => $Params{Mapping});
-	    my $u = $s;
-	    $@ = '';
-	    eval {
-		$u = $csetobj->decode($u, 0);
-	    } if MIME::Charset::USE_ENCODE;
-	    if ($@ or $u =~ $UNSAFE) {
-		$enc = $Params{Encoding};
-	    } else {
-		($cset, $enc) = ("US-ASCII", undef);
-		$csetobj = MIME::Charset->new($cset, Mapping => "STANDARD");
-	    }
+	($s, $cset, $enc) =
+	    $obj->header_encode($s,
+				Detect7bit => $Params{Detect7bit},
+				Replacement => $Params{Replacement},
+				Encoding => $Params{Encoding});
+	# pure ASCII
+	if ($cset eq "US-ASCII" and !$enc and $s =~ /$UNSAFEASCII/) {
+	    # pure ASCII with unsafe sequences should be encoded
+	    $cset = $csetobj->output_charset ||
+		$charsetobj->output_charset ||
+		$ascii->output_charset;
+	    $csetobj = MIME::Charset->new($cset,
+					  Mapping => $Params{Mapping});
+	    $enc = $csetobj->header_encoding || 'Q';
+	} else {
+	    $csetobj = MIME::Charset->new($cset,
+					  Mapping => $Params{Mapping});
 	}
 
-	# Now no charset transformations are needed.
+	# Now no charset translations are needed.
 	$csetobj->encoder($csetobj);
 
 	# Concatenate adjacent ``words'' so that multibyte sequences will
@@ -754,8 +787,8 @@ sub encode_mimewords  {
 		($lastenc || "") eq ($enc || "")) {
 		$triplets[-1]->[0] .= $s;
 		next;
-	    } elsif (!$lastenc and $enc and $last !~ /[\t ]$/) {
-		if ($last =~ /^(.*)([\t ])([$PRINTABLE]+)$/s) {
+	    } elsif (!$lastenc and $enc and $last !~ /[\r\n\t ]$/) {
+		if ($last =~ /^(.*)([\r\n\t ])([$PRINTABLE]+)$/s) {
 		    $triplets[-1]->[0] = $1.$2;
 		    $s = $3.$s;
 		} elsif ($lastcsetobj->as_string eq "US-ASCII") {
@@ -764,8 +797,8 @@ sub encode_mimewords  {
 		    $triplets[-1]->[2] = $csetobj;
 		    next;
 		}
-	    } elsif ($lastenc and !$enc and $s !~ /^[\t ]/) {
-		if ($s =~ /^([$PRINTABLE]+)([\t ])(.*)$/s) {
+	    } elsif ($lastenc and !$enc and $s !~ /^[\r\n\t ]/) {
+		if ($s =~ /^([$PRINTABLE]+)([\r\n\t ])(.*)$/s) {
 		    $triplets[-1]->[0] .= $1;
 		    $s = $2.$3;
 		} elsif ($csetobj->as_string eq "US-ASCII") {
@@ -778,8 +811,8 @@ sub encode_mimewords  {
     }
 
     # chop leading FWS
-    while ($triplets[0]->[0] =~ s/^[\r\n\t ]+//) {
-	shift @triplets unless $triplets[0]->[0];
+    while (scalar(@triplets) and $triplets[0]->[0] =~ s/^[\r\n\t ]+//) {
+	shift @triplets unless length($triplets[0]->[0]);
     }
 
     # Split long ``words''.
@@ -788,23 +821,21 @@ sub encode_mimewords  {
     foreach (@triplets) {
 	my ($s, $enc, $csetobj) = @$_;
 
-	if ($enc and $restlen < $csetobj->encoded_header_len("\a", $enc) or
-	    !$enc and $restlen < length($s) - ($s =~ /^[\t ]/? 1: 0)) {
-	    $restlen = $maxrestlen;
-	}
-
 	push @splitted, &_split($s, $enc, $csetobj, $restlen, $maxrestlen);
 	my ($last, $lastenc, $lastcsetobj) = @{$splitted[-1]};
+	my $lastlen;
 	if ($lastenc) {
-	    $restlen -= $lastcsetobj->encoded_header_len($last, $lastenc);
+	    $lastlen = $lastcsetobj->encoded_header_len($last, $lastenc);
 	} else {
-	    $restlen -= length($last); # FIXME: Sometimes estimated longer
+	    $lastlen = length($last);
 	}
+	$restlen -= $lastlen; # FIXME: Sometimes estimated longer
+	$restlen = $maxrestlen if $restlen <= 1;
     }
 
     # Do encoding.
     my @lines;
-    my $linelen = $firstlinelen;
+    $restlen = $firstlinelen;
     foreach (@splitted) {
 	my ($str, $encoding, $charsetobj) = @$_;
 	next unless length($str);
@@ -813,22 +844,22 @@ sub encode_mimewords  {
 	if (!$encoding) {
 	    $s = $str;
 	} else {
-	    $s = &encode_mimeword($str, $encoding, $charsetobj->as_string);
+	    $s = encode_mimeword($str, $encoding, $charsetobj);
 	}
 
-	my $spc = (scalar(@lines) and $lines[-1] =~ /[\t ]$/ or
-		   $s =~ /^[\t ]/)? '': ' ';
+	my $spc = (scalar(@lines) and $lines[-1] =~ /[\r\n\t ]$/ or
+		   $s =~ /^[\r\n\t ]/)? '': ' ';
 	if (!scalar(@lines)) {
 	    push @lines, $s;
-	} elsif (length($lines[-1].$spc.$s) <= $linelen) {
+	} elsif (length($lines[-1].$spc.$s) <= $restlen) {
 	    $lines[-1] .= $spc.$s;
 	} else {
 	    if ($lines[-1] =~ s/([\r\n\t ]+)$//) {
 		$s = $1.$s;
 	    }
-	    $s =~ s/^[\t ]//; # strip only one WSP replaced by that of FWS
+	    $s =~ s/^[\r\n]*[\t ]//; # strip only one WSP replaced by FWS
 	    push @lines, $s;
-	    $linelen = $maxrestlen;
+	    $restlen = $maxrestlen;
 	}
     }
 
@@ -866,10 +897,17 @@ sub _split {
 	    last;
 	}
 	$ustr = $str;
-	if (MIME::Charset::USE_ENCODE) {
+	if (!(is_utf8($ustr) or $ustr =~ /$WIDECHAR/) and
+	    MIME::Charset::USE_ENCODE) {
 	    $ustr = $charset->decode($ustr);
 	}
 	($first, $str) = &_clip_unsafe($ustr, $encoding, $charset, $restlen);
+	# retry splitting if failed
+	if ($first and !$str and
+	    $maxrestlen < $charset->encoded_header_len($first, $encoding)) {
+	    ($first, $str) = &_clip_unsafe($ustr, $encoding, $charset,
+					   $maxrestlen);
+	}
 	push @splitted, [$first, $encoding, $charset];
 	$restlen = $maxrestlen;
     }
@@ -879,9 +917,7 @@ sub _split {
 # _split_ascii RAW, ROOM_OF_FIRST_LINE, MAXRESTLEN
 #     Private: used by encode_mimewords() to split an US-ASCII string into
 #     (encoded or non-encoded) words.
-#     Returns an array of arrayrefs [SUBSTRING, ENCODING, "US-ASCII"],
-#     where ENCODING is either undef or (if any unsafe sequences are
-#     included) "Q".
+#     Returns an array of arrayrefs [SUBSTRING, undef, "US-ASCII"].
 sub _split_ascii {
     my $s = shift;
     my $restlen = shift;
@@ -891,12 +927,6 @@ sub _split_ascii {
     my @splitted;
     my $ascii = MIME::Charset->new("US-ASCII", Mapping => 'STANDARD');
     foreach my $line (split(/(?:[\t ]*[\r\n]+)+/, $s)) {
-	if (length($line) <= $restlen and $line !~ /=\?|$UNSAFE/) {
-	    push @splitted, [$line, undef, $ascii];
-	    $restlen = $maxrestlen;
-	    next;
-	}
-
         my $spc = '';
 	foreach my $word (split(/([\t ]+)/, $line)) {
 	    next unless scalar(@splitted) or $word; # skip first garbage
@@ -905,44 +935,22 @@ sub _split_ascii {
 		next;
 	    }
 
-	    my $enc = ($word =~ /=\?|$UNSAFE/)? "Q": undef;
+	    my $cont = $spc.$word;
+	    my $elen = length($cont);
+	    next unless $elen;
 	    if (scalar(@splitted)) {
-		my ($last, $lastenc, $lastcsetobj) = @{$splitted[-1]};
-		my ($elen, $cont, $appe);
-
 		# Concatenate adjacent words so that encoded-word and
 		# unencoded text will adjoin with separating whitespace.
-		if (!$lastenc and !$enc) {
-		    $elen = length($spc.$word);
-		    ($cont, $appe) = ($spc.$word, "");
-		} elsif (!$lastenc and $enc) {
-		    $elen = length($spc) +
-			$ascii->encoded_header_len($word, "Q");
-		    ($cont, $appe) = ($spc, $word);
-		} elsif ($lastenc and !$enc) {
-		    $elen = length($spc.$word);
-		    ($cont, $appe) = ("", $spc.$word);
-		} else {
-		    $elen = $ascii->encoded_header_len($spc.$word, "Q") - 15;
-		    ($cont, $appe) = ($spc.$word, "");
-		}
 		if ($elen <= $restlen) {
-		    $splitted[-1]->[0] .= $cont if $cont;
-		    push @splitted, [$appe, $enc, $ascii] if $appe;
+		    $splitted[-1]->[0] .= $cont;
 		    $restlen -= $elen;
 		} else {
-		    push @splitted, [$cont, $lastenc, $ascii] if $cont;
-		    push @splitted, [$appe, $enc, $ascii] if $appe;
+		    push @splitted, [$cont, undef, $ascii];
 		    $restlen = $maxrestlen - $elen;
-		    $restlen -= 15 if $lastenc and $enc;
 		}
 	    } else {
-		push @splitted, [$spc.$word, $enc, $ascii];
-		if ($enc) {
-		    $restlen -= $ascii->encoded_header_len($spc.$word, "Q");
-		} else {
-		    $restlen -= length($spc.$word);
-		}
+		push @splitted, [$cont, undef, $ascii];
+		$restlen -= $elen;
 	    }
 	    $spc = '';
 	}
@@ -950,7 +958,7 @@ sub _split_ascii {
 	    if (scalar(@splitted)) {
 		$splitted[-1]->[0] .= $spc;
 		$restlen -= length($spc);
-	    } else { # NOTREACHED
+	    } else { # only WSPs
 		push @splitted, [$spc, undef, $ascii];
 		$restlen = $maxrestlen - length($spc);
 	    }
